@@ -3,6 +3,7 @@
  *  oFono - Open Source Telephony
  *
  *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2010 ST-Ericsson AB.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -27,6 +28,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <errno.h>
 
 #include <glib.h>
 
@@ -38,6 +40,7 @@
 #include "gatresult.h"
 
 #include "atmodem.h"
+#include "vendor.h"
 
 static const char *cgreg_prefix[] = { "+CGREG:", NULL };
 static const char *cgdcont_prefix[] = { "+CGDCONT:", NULL };
@@ -45,6 +48,7 @@ static const char *none_prefix[] = { NULL };
 
 struct gprs_data {
 	GAtChat *chat;
+	unsigned int vendor;
 };
 
 static void at_cgatt_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -53,7 +57,6 @@ static void at_cgatt_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	ofono_gprs_cb_t cb = cbd->cb;
 	struct ofono_error error;
 
-	dump_response("cgatt_cb", ok, result);
 	decode_at_error(&error, g_at_result_final_response(result));
 
 	cb(&error, cbd->data);
@@ -66,18 +69,13 @@ static void at_gprs_set_attached(struct ofono_gprs *gprs, int attached,
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 
-	if (!cbd)
-		goto error;
-
-	sprintf(buf, "AT+CGATT=%i", attached ? 1 : 0);
+	snprintf(buf, sizeof(buf), "AT+CGATT=%i", attached ? 1 : 0);
 
 	if (g_at_chat_send(gd->chat, buf, none_prefix,
 				at_cgatt_cb, cbd, g_free) > 0)
 		return;
 
-error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -87,23 +85,23 @@ static void at_cgreg_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct cb_data *cbd = user_data;
 	ofono_gprs_status_cb_t cb = cbd->cb;
 	struct ofono_error error;
-	int status, lac, ci, tech;
+	int status;
+	struct gprs_data *gd = cbd->user;
 
-	dump_response("at_cgreg_cb", ok, result);
 	decode_at_error(&error, g_at_result_final_response(result));
 
 	if (!ok) {
-		cb(&error, -1, -1, -1, -1, cbd->data);
+		cb(&error, -1, cbd->data);
 		return;
 	}
 
 	if (at_util_parse_reg(result, "+CGREG:", NULL, &status,
-				&lac, &ci, &tech) == FALSE) {
-		CALLBACK_WITH_FAILURE(cb, -1, -1, -1, -1, cbd->data);
+				NULL, NULL, NULL, gd->vendor) == FALSE) {
+		CALLBACK_WITH_FAILURE(cb, -1, cbd->data);
 		return;
 	}
 
-	cb(&error, status, lac, ci, tech, cbd->data);
+	cb(&error, status, cbd->data);
 }
 
 static void at_gprs_registration_status(struct ofono_gprs *gprs,
@@ -113,32 +111,47 @@ static void at_gprs_registration_status(struct ofono_gprs *gprs,
 	struct gprs_data *gd = ofono_gprs_get_data(gprs);
 	struct cb_data *cbd = cb_data_new(cb, data);
 
-	if (!cbd)
-		goto error;
+	cbd->user = gd;
+
+	switch (gd->vendor) {
+	case OFONO_VENDOR_GOBI:
+		/*
+		 * Send *CNTI=0 to find out the current tech, it will be
+		 * intercepted in gobi_cnti_notify in network registration
+		 */
+		g_at_chat_send(gd->chat, "AT*CNTI=0", none_prefix,
+				NULL, NULL, NULL);
+		break;
+	case OFONO_VENDOR_NOVATEL:
+		/*
+		 * Send $CNTI=0 to find out the current tech, it will be
+		 * intercepted in nw_cnti_notify in network registration
+		 */
+		g_at_chat_send(gd->chat, "AT$CNTI=0", none_prefix,
+				NULL, NULL, NULL);
+		break;
+	}
 
 	if (g_at_chat_send(gd->chat, "AT+CGREG?", cgreg_prefix,
 				at_cgreg_cb, cbd, g_free) > 0)
 		return;
 
-error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
 
-	CALLBACK_WITH_FAILURE(cb, -1, -1, -1, -1, data);
+	CALLBACK_WITH_FAILURE(cb, -1, data);
 }
 
 static void cgreg_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_gprs *gprs = user_data;
-	int status, lac, ci, tech;
-
-	dump_response("cgreg_notify", TRUE, result);
+	int status;
+	struct gprs_data *gd = ofono_gprs_get_data(gprs);
 
 	if (at_util_parse_reg_unsolicited(result, "+CGREG:", &status,
-				&lac, &ci, &tech) == FALSE)
+				NULL, NULL, NULL, gd->vendor) == FALSE)
 		return;
 
-	ofono_gprs_status_notify(gprs, status, lac, ci, tech);
+	ofono_gprs_status_notify(gprs, status);
 }
 
 static void cgev_notify(GAtResult *result, gpointer user_data)
@@ -162,6 +175,51 @@ static void cgev_notify(GAtResult *result, gpointer user_data)
 	}
 }
 
+static void xdatastat_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_gprs *gprs = user_data;
+	GAtResultIter iter;
+	int stat;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "+XDATASTAT:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &stat))
+
+	DBG("stat %d", stat);
+
+	switch (stat) {
+	case 0:
+		ofono_gprs_suspend_notify(gprs, GPRS_SUSPENDED_UNKNOWN_CAUSE);
+		break;
+	case 1:
+		ofono_gprs_resume_notify(gprs);
+		break;
+	}
+}
+
+static void cpsb_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_gprs *gprs = user_data;
+	GAtResultIter iter;
+	gint bearer;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "+CPSB:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, NULL))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &bearer))
+		return;
+
+	ofono_gprs_bearer_notify(gprs, bearer);
+}
+
 static void gprs_initialized(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_gprs *gprs = user_data;
@@ -169,7 +227,20 @@ static void gprs_initialized(gboolean ok, GAtResult *result, gpointer user_data)
 
 	g_at_chat_register(gd->chat, "+CGEV:", cgev_notify, FALSE, gprs, NULL);
 	g_at_chat_register(gd->chat, "+CGREG:", cgreg_notify,
-				FALSE, gprs, NULL);
+						FALSE, gprs, NULL);
+	g_at_chat_register(gd->chat, "+CPSB:", cpsb_notify, FALSE, gprs, NULL);
+
+	g_at_chat_send(gd->chat, "AT+CPSB=1", none_prefix, NULL, NULL, NULL);
+
+	switch (gd->vendor) {
+	case OFONO_VENDOR_IFX:
+		/* Register for GPRS suspend notifications */
+		g_at_chat_register(gd->chat, "+XDATASTAT:", xdatastat_notify,
+						FALSE, gprs, NULL);
+		g_at_chat_send(gd->chat, "AT+XDATASTAT=1", none_prefix,
+						NULL, NULL, NULL);
+		break;
+	}
 
 	ofono_gprs_register(gprs);
 }
@@ -185,8 +256,6 @@ static void at_cgreg_test_cb(gboolean ok, GAtResult *result,
 	int cgreg2 = 0;
 	const char *cmd;
 
-	dump_response("cgreg_read_cb", ok, result);
-
 	if (!ok)
 		goto error;
 
@@ -216,8 +285,24 @@ static void at_cgreg_test_cb(gboolean ok, GAtResult *result,
 
 	g_at_chat_send(gd->chat, cmd, none_prefix, NULL, NULL, NULL);
 	g_at_chat_send(gd->chat, "AT+CGAUTO=0", none_prefix, NULL, NULL, NULL);
-	g_at_chat_send(gd->chat, "AT+CGEREP=2,1", none_prefix,
+
+	switch (gd->vendor) {
+	case OFONO_VENDOR_MBM:
+		/* Ericsson MBM and ST-E modems don't support AT+CGEREP=2,1 */
+		g_at_chat_send(gd->chat, "AT+CGEREP=1,0", none_prefix,
 			gprs_initialized, gprs, NULL);
+		break;
+	case OFONO_VENDOR_NOKIA:
+		/* Nokia data cards don't support AT+CGEREP=1,0 either */
+		g_at_chat_send(gd->chat, "AT+CGEREP=1", none_prefix,
+			gprs_initialized, gprs, NULL);
+		break;
+	default:
+		g_at_chat_send(gd->chat, "AT+CGEREP=2,1", none_prefix,
+			gprs_initialized, gprs, NULL);
+		break;
+	}
+
 	return;
 
 error:
@@ -287,12 +372,16 @@ static int at_gprs_probe(struct ofono_gprs *gprs,
 	GAtChat *chat = data;
 	struct gprs_data *gd;
 
-	gd = g_new0(struct gprs_data, 1);
-	gd->chat = chat;
+	gd = g_try_new0(struct gprs_data, 1);
+	if (gd == NULL)
+		return -ENOMEM;
+
+	gd->chat = g_at_chat_clone(chat);
+	gd->vendor = vendor;
 
 	ofono_gprs_set_data(gprs, gd);
 
-	g_at_chat_send(chat, "AT+CGDCONT=?", cgdcont_prefix,
+	g_at_chat_send(gd->chat, "AT+CGDCONT=?", cgdcont_prefix,
 			at_cgdcont_test_cb, gprs, NULL);
 
 	return 0;
@@ -303,6 +392,8 @@ static void at_gprs_remove(struct ofono_gprs *gprs)
 	struct gprs_data *gd = ofono_gprs_get_data(gprs);
 
 	ofono_gprs_set_data(gprs, NULL);
+
+	g_at_chat_unref(gd->chat);
 	g_free(gd);
 }
 
@@ -311,15 +402,15 @@ static struct ofono_gprs_driver driver = {
 	.probe			= at_gprs_probe,
 	.remove			= at_gprs_remove,
 	.set_attached		= at_gprs_set_attached,
-	.registration_status	= at_gprs_registration_status,
+	.attached_status	= at_gprs_registration_status,
 };
 
-void at_gprs_init()
+void at_gprs_init(void)
 {
 	ofono_gprs_driver_register(&driver);
 }
 
-void at_gprs_exit()
+void at_gprs_exit(void)
 {
 	ofono_gprs_driver_unregister(&driver);
 }
